/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19/05/2007
   derniere modification : 21/07/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEMatrix.h>
#include <SCE/utils/SCEResources.h>
#include <SCE/core/SCECore.h>
#include <SCE/core/SCECMatrix.h>

#include <SCE/interface/SCEMaterial.h>

/* ajoute le 13/01/2008 */
/* revise le 19/08/2008 */
static void SCE_Material_InitMatTex (SCE_SMatTex *t)
{
    t->tex = NULL;
    t->actived = SCE_FALSE;
    t->canfree = SCE_FALSE;
    t->unit = 0;
}
/* ajoute le 15/03/2008 */
/* revise le 04/04/2008 */
static void SCE_Material_InitMatShd (SCE_SMatShd *s)
{
    int i;
    s->shd = NULL;
    s->actived = SCE_FALSE;
    s->canfree = SCE_FALSE;
    for (i=0; i<SCE_NUM_LIGHT_PARAMS; i++)
        s->lightindex[i] = -1; /* TODO: incompatible avec Cg */
}

/* ajoute le 20/07/2008 */
static void SCE_Material_InitStates (SCE_SMaterialStates *states)
{
    states->use_ps = SCE_FALSE;
    states->use_sl = SCE_FALSE;
}

/* ajoute le 13/01/2008 */
/* revise le 19/08/2008 */
static void SCE_Material_Init (SCE_SMaterial *m)
{
    unsigned int i;
    m->mat = NULL;
    m->maps = NULL;
    SCE_Material_InitMatShd (&m->shader);
    SCE_CInitPointSprite (&m->ps);
    SCE_Material_InitStates (&m->states);
    m->cubemapping = SCE_FALSE;
    m->cubemap_unit = 0;
    for (i=0; i<SCE_MAX_TEXTURE_UNITS; i++)
        SCE_Matrix4_Identity (m->matrices[i]);
}

/* ajoute le 13/01/2008 */
static SCE_SMatTex* SCE_Material_CreateMatTex (void)
{
    SCE_SMatTex *t = SCE_malloc (sizeof *t);
    if (t)
        SCE_Material_InitMatTex (t);
    else
        Logger_LogSrc ();
    return t;
}

/* ajoute le 13/01/2008 */
static void SCE_Material_DeleteMatTex (void *tex)
{
    if (tex)
    {
        SCE_SMatTex *t = tex;
        if (t->canfree)
            SCE_Texture_Delete (t->tex);
        SCE_free (t);
    }
}

/* revise le 14/03/2008 */
SCE_SMaterial* SCE_Material_Create (void)
{
    SCE_SMaterial *mat = NULL;

    SCE_btstart ();
    mat = SCE_malloc (sizeof *mat);
    if (!mat)
        Logger_LogSrc ();
    else
    {
        SCE_Material_Init (mat);

        if (!(mat->mat = SCE_CCreateMaterial ()))
        {
            SCE_free (mat);
            Logger_LogSrc ();
            SCE_btend ();
            return NULL;
        }

        if (!(mat->maps = SCE_List_Create (SCE_Material_DeleteMatTex)))
        {
            SCE_Material_Delete (mat), mat = NULL;
            Logger_LogSrc ();
        }

    }
    SCE_btend ();
    return mat;
}

/* revise le 19/08/2008 */
void SCE_Material_Delete (SCE_SMaterial *mat)
{
    SCE_btstart ();
    if (mat)
    {
        if (mat->shader.canfree)
            SCE_Shader_Delete (mat->shader.shd);
        SCE_List_Delete (mat->maps);
        SCE_CDeleteMaterial (mat->mat);
        SCE_free (mat);
    }
    SCE_btend ();
}


/* ajoute le 13/01/2008 */
SCE_CMaterial* SCE_Material_GetCMaterial (SCE_SMaterial *mat)
{
    return mat->mat;
}


/* ajoute le 13/01/2008 */
void SCE_Material_SetColor (SCE_SMaterial *mat, SCEenum type,
                            float r, float g, float b, float a)
{
    SCE_CSetMaterialColor (mat->mat, type, r, g, b, a);
}
void SCE_Material_SetColorv (SCE_SMaterial *mat, SCEenum type, float *color)
{
    SCE_CSetMaterialColorv (mat->mat, type, color);
}

float* SCE_Material_GetColor (SCE_SMaterial *mat, SCEenum type)
{
    return SCE_CGetMaterialColor (mat->mat, type);
}
void SCE_Material_GetColorv (SCE_SMaterial *mat, SCEenum type, float *color)
{
    SCE_CGetMaterialColorv (mat->mat, type, color);
}


/* ajoute le 25/03/2008 */
void SCE_Material_ActivatePointSprite (SCE_SMaterial *mat, int actived)
{
    mat->states.use_ps = actived;
}
void SCE_Material_EnablePointSprite (SCE_SMaterial *mat)
{
    mat->states.use_ps = SCE_TRUE;
}
void SCE_Material_DisablePointSprite (SCE_SMaterial *mat)
{
    mat->states.use_ps = SCE_FALSE;
}
/* ajoute le 25/03/2008 */
void SCE_Material_SetPointSpriteModel (SCE_SMaterial *mat, SCE_CPointSprite *ps)
{
    memcpy (&mat->ps, ps, sizeof *ps);
}
SCE_CPointSprite* SCE_Material_GetPointSpriteModel (SCE_SMaterial *mat)
{
    return &mat->ps;
}


/* ajoute le 24/03/2008 */
void SCE_Material_ActivateBlending (SCE_SMaterial *mat, int use)
{
    SCE_CActiveMaterialBlending (mat->mat, use);
}
void SCE_Material_EnableBlending (SCE_SMaterial *mat)
{
    SCE_CEnableMaterialBlending (mat->mat);
}
void SCE_Material_DisableBlending (SCE_SMaterial *mat)
{
    SCE_CDisableMaterialBlending (mat->mat);
}
/* ajoute le 24/03/2008 */
void SCE_Material_SetBlending (SCE_SMaterial *mat, SCEenum src, SCEenum dst)
{
    SCE_CSetMaterialBlending (mat->mat, src, dst);
}


/* ajoute le 05/07/2008 */
void SCE_Material_ActivateShaderLighting (SCE_SMaterial *mat, int use)
{
    mat->states.use_sl = use;
}
void SCE_Material_EnableShaderLighting (SCE_SMaterial *mat)
{
    mat->states.use_sl = SCE_TRUE;
}
void SCE_Material_DisableShaderLighting (SCE_SMaterial *mat)
{
    mat->states.use_sl = SCE_FALSE;
}


/* ajoute le 20/07/2008 */
void SCE_Material_EnableCubeMapping (SCE_SMaterial *mat, unsigned int unit)
{
    mat->cubemapping = SCE_TRUE;
    mat->cubemap_unit = unit;
}
void SCE_Material_DisableCubeMapping (SCE_SMaterial *mat)
{
    mat->cubemapping = SCE_FALSE;
}


/* ajoute le 19/08/2008 */
float* SCE_Material_GetTextureMatrix (SCE_SMaterial *mat, unsigned int unit)
{
    return mat->matrices[unit];
}


/* ajoute le 12/03/2008 */
SCE_STexture* SCE_Material_GetTexture (SCE_SMaterial *mat, unsigned int id)
{
    SCE_SMatTex *tex = SCE_List_GetData (SCE_List_GetIterator (mat->maps, id));
    return tex->tex;
}
/* ajoute le 12/03/2008 */
SCE_SShader* SCE_Material_GetShader (SCE_SMaterial *mat)
{
    return mat->shader.shd;
}

/* revise le 06/04/2008 */
int SCE_Material_AddTexture (SCE_SMaterial *mat, SCEuint unit,
                             SCE_STexture *tex, int canfree)
{
    SCE_SMatTex *t = NULL;

    SCE_btstart ();
    /* TODO: placer ca dans une fonction */
    if (unit >= SCE_CGetMaxTextureUnits ())
    {
        Logger_Log (SCE_INVALID_ARG);
        Logger_LogMsg ("the %dth texture unit is not"
                       "available on your hardware", unit);
        return SCE_ERROR;
    }

#define SCE_matassert(c)\
    if (c)\
    {\
        SCE_Material_DeleteMatTex (t);\
        Logger_LogSrc ();\
        SCE_btend ();\
        return SCE_ERROR;\
    }

    t = SCE_Material_CreateMatTex ();
    SCE_matassert (!t)
    SCE_matassert (SCE_List_AddBack (mat->maps, t) < 0)

    t->tex = tex;
    t->canfree = canfree;
    t->unit = unit;
    t->actived = SCE_TRUE;

#undef SCE_matassert

    SCE_btend ();
    return SCE_OK;
}

/* revise le 14/03/2008 */
void SCE_Material_SetShader (SCE_SMaterial *mat, SCE_SShader *shd, int canfree)
{
    mat->shader.shd = shd;
    mat->shader.canfree = canfree;
    mat->shader.actived = SCE_TRUE;
    /* recherche des indices des parametres de lumiere */
    SCE_Light_MakeIndices (shd, mat->shader.lightindex);
}



/* ajoute le 13/01/2008 */
/* revise le 14/01/2008 */
int SCE_Material_LoadTextureArg (SCE_SMaterial *mat, SCEenum type,
                                 SCEuint unit, int use_mipmap, va_list args)
{
    SCE_STexture *tex = NULL;

    SCE_btstart ();
#define SCE_matassert(c)\
    if (c)\
    {\
        SCE_Texture_Delete (tex);\
        Logger_LogSrc ();\
        SCE_btend ();\
        return SCE_ERROR;\
    }

    tex = SCE_Texture_LoadArg (type, 0, 0, 0, args);
    SCE_matassert (!tex)
    SCE_matassert (SCE_Texture_Build (tex, use_mipmap) < 0)
    SCE_matassert (SCE_Material_AddTexture (mat, unit, tex, SCE_TRUE) < 0)

#undef SCE_matassert

    SCE_btend ();
    return SCE_OK;
}
/* revise le 13/01/2008 */
int SCE_Material_LoadTexture (SCE_SMaterial *mat, SCEenum type, SCEuint unit,
                              int use_mipmap, ...)
{
    va_list args;
    int r;
    va_start (args, use_mipmap);
    r = SCE_Material_LoadTextureArg (mat, type, unit, use_mipmap, args);
    va_end (args);
    return r;
}


int SCE_Material_LoadShader (SCE_SMaterial *mat, const char *vs, const char *ps)
{
    SCE_SShader *shd = NULL;

    SCE_btstart ();
    shd = SCE_Shader_CreateFromFile (vs, ps);
    if (!shd)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    if (SCE_Shader_Build (shd) < 0)
    {
        SCE_Shader_Delete (shd);
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    SCE_Material_SetShader (mat, shd, SCE_TRUE);

    SCE_btend ();
    return SCE_OK;
}

#if 0
/* ajoute le 14/06/2007 */
/* revise le 19/06/2007 */
void SCE_Material_RemoveTexture (SCE_SMaterial *mat, int type, int unit)
{
    
}

/* ajoute le 14/06/2007 */
/* revise le 18/06/2007 */
void SCE_Material_RemoveShader (SCE_SMaterial *mat)
{
    
}

/* ajoute le 14/06/2007 */
/* revise le 19/06/2007 */
void SCE_Material_ActivateTexture (SCE_SMaterial *mat, int unit, int state)
{
    
}
#endif

/* ajoute le 19/08/2008 */
void SCE_Material_ActivateShader (SCE_SMaterial *mat, int actived)
{
    mat->shader.actived = actived;
}

/* ajoute le 19/08/2008 */
static void SCE_Material_SetupCubeMapping (void)
{
    SCE_TMatrix4 mat;
    /* activation du cubemapping */
    SCE_CSetState3 (GL_TEXTURE_GEN_S, GL_TEXTURE_GEN_T,
                    GL_TEXTURE_GEN_R, SCE_TRUE);
    SCE_CSetTextureGen_ (GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
    SCE_CSetTextureGen_ (GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
    SCE_CSetTextureGen_ (GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
    SCE_CGetMatrix (SCE_MAT_MODELVIEW, mat); /* TODO: BOUH QUE C'EST LAID
                                              recuperer la matrice comme ca */
    SCE_Matrix4_Inverse (mat);
    mat[3] = mat[7] = mat[11] = 0.0f;
    mat[15] = 1.0f;
    SCE_CMultMatrix (mat);
}

/* ajoute le 21/07/2008 */
static void SCE_Material_ClearTextures (void)
{
    unsigned int i;
    SCE_CSetActiveMatrix (SCE_MAT_TEXTURE);
    for (i=0; i<SCE_CGetMaxTextureUnits (); i++)
    {
        SCE_CSetActiveTextureUnit (i);
        SCE_Texture_Use (NULL, i);
        SCE_CPopMatrix ();
    }
}

/* ajoute le 19/08/2008 */
static void SCE_Material_SetupTextureMatrices (SCE_SMaterial *mat)
{
    unsigned int i;
    SCE_CSetActiveMatrix (SCE_MAT_TEXTURE);
    for (i=0; i<SCE_CGetMaxTextureUnits (); i++)
    {
        SCE_CSetActiveTextureUnit (i);
        SCE_CPushMatrix ();
        /*SCE_CLoadIdentityMatrix ();*/ /* useless call */
        SCE_CLoadMatrix (mat->matrices[i]);
    }
}


/* revise le 19/08/2008 */
void SCE_Material_Use (SCE_SMaterial *mat)
{
    static int matinuse = SCE_FALSE;
    /* harhar */
    SCE_CSetState3 (GL_TEXTURE_GEN_S, GL_TEXTURE_GEN_T,
                    GL_TEXTURE_GEN_R, SCE_FALSE);

    if (mat)
    {
        SCE_SMatTex *t = NULL;
        SCE_SListIterator *it = NULL;

        if (matinuse)
            SCE_Material_ClearTextures ();
        SCE_Material_SetupTextureMatrices (mat);

        SCE_List_ForEach (it, mat->maps)
        {
            t = SCE_List_GetData (it);
            if (t->actived)
            {
                SCE_Texture_Use (t->tex, t->unit);
                if (mat->cubemapping && t->unit == mat->cubemap_unit)
                    SCE_Material_SetupCubeMapping ();
            }
        }

        if (mat->shader.actived)
        {
            SCE_Shader_Use (mat->shader.shd);
            SCE_Light_Apply (mat->shader.lightindex);
        }
        else
            SCE_Shader_Use (NULL);

        SCE_CUseMaterial (mat->mat);
        SCE_CBindPointSprite ((mat->states.use_ps) ? &mat->ps : NULL);
        matinuse = SCE_TRUE;
    }
    else
    {
        SCE_CBindPointSprite (NULL);
        SCE_CUseMaterial (NULL);
        SCE_Shader_Use (NULL);
        SCE_Material_ClearTextures ();
        matinuse = SCE_FALSE;
    }
    SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
}
